export class HighlightWithInTextarea {
    private isGenerated: boolean = false;

    private readonly prefixIdForNewElement?: string = undefined;
    private readonly targetElement: HTMLTextAreaElement = new HTMLTextAreaElement();
    private highlightsElement: HTMLElement = new HTMLElement();
    private backdropElement: HTMLElement = new HTMLElement();
    private containerElement: HTMLElement = new HTMLElement();
    private readonly highlightConfig?: HighlightConfig = undefined;
    private userBrowser?: BrowserType = undefined;

    constructor(targetElement: HTMLTextAreaElement, config: HighlightConfig) {
        this.prefixIdForNewElement = `highlight-${this.generateUUIDv4WithoutHyphens()}`;

        this.targetElement = targetElement;

        if (config && config.highlight) {
            this.highlightConfig = config;
            this.generate();
        } else {
            console.error('valid config object not provided');
            throw new Error('valid config object not provided')
        }
    }

    private generateUUIDv4WithoutHyphens() {
        return 'xxxxxxxxxxxx4xxxyxxxxxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
            const r = (Math.random() * 16) | 0,
                v = c === 'x' ? r : (r & 0x3) | 0x8;
            return v.toString(16);
        });
    }

    private getType(instance?: unknown): HighlightType {
        let type = typeof instance;
        if (!instance) {
            return 'falsely';
        } else if (Array.isArray(instance)) {
            if (instance.length === 2 && typeof instance[0] === 'number' && typeof instance[1] === 'number') {
                return 'range';
            } else {
                return 'array';
            }
        } else if (type === 'object') {
            if (instance instanceof RegExp) {
                return 'regexp';
            } else if (instance.hasOwnProperty('highlight')) {
                return 'custom';
            }
        } else if (type === 'function' || type === 'string') {
            return type;
        }

        return 'other';
    }

    private generate() {
        this?.targetElement?.classList.add(this.prefixIdForNewElement + '-input', this.prefixIdForNewElement + '-content');
        this?.targetElement?.addEventListener('input', this.handleInput.bind(this));
        this?.targetElement?.addEventListener('scroll', this.handleScroll.bind(this));

        this.highlightsElement = document.createElement('div');
        this.highlightsElement.className = this.prefixIdForNewElement + '-highlights ' + this.prefixIdForNewElement + '-content';

        this.backdropElement = document.createElement('div');
        this.backdropElement.className = this.prefixIdForNewElement + '-backdrop';
        this.backdropElement.appendChild(this.highlightsElement);

        this.containerElement = document.createElement('div');
        this.containerElement.className = this.prefixIdForNewElement + '-container';
        this.containerElement.appendChild(this.backdropElement);
        this.containerElement.appendChild(this.targetElement); // targetElement을 containerElement 내로 이동
        this.containerElement.addEventListener('scroll', this.blockContainerScroll.bind(this));

        this.userBrowser = this.detectBrowser();
        switch (this.userBrowser) {
            case 'firefox':
                this.fixFirefox();
                break;
            case 'ios':
                this.fixIOS();
                break;
        }

        // 플러그인 함수가 이를 확인할 수 있도록 설정
        this.isGenerated = true;

        // 기존 입력을 하이라이트하기 위해 input 이벤트를 트리거
        this.handleInput();


        document.body.appendChild(this.containerElement);
    }

    private detectBrowser(): BrowserType {
        let ua = window.navigator.userAgent.toLowerCase();
        if (ua.indexOf('firefox') !== -1) {
            return 'firefox';
        } else if (ua.includes('msie') || ua.includes('trident/7') || ua.includes('edge')) {
            return 'ie';
        } else if (ua.includes('ipad') || ua.includes('iphone') || (ua.includes('ipod') && !ua.includes('windows phone'))) {
            return 'ios';
        } else {
            return 'other';
        }
    };

    private fixFirefox() {
        // highlights div에서 패딩 및 보더 픽셀을 가져옵니다.
        let padding = window.getComputedStyle(this.highlightsElement);
        let border = window.getComputedStyle(this.highlightsElement);

        // 패딩 및 보더 픽셀 값을 0으로 설정합니다.
        if (this.highlightsElement) {
            this.highlightsElement.style.padding = '0';
            this.highlightsElement.style.borderWidth = '0';
        }

        if (this.backdropElement) {
            // backdrop div에 패딩 및 보더 픽셀을 추가합니다.
            this.backdropElement.style.marginTop = `+=${padding.getPropertyValue('padding-top')}`;
            this.backdropElement.style.marginRight = `+=${padding.getPropertyValue('padding-right')}`;
            this.backdropElement.style.marginBottom = `+=${padding.getPropertyValue('padding-bottom')}`;
            this.backdropElement.style.marginLeft = `+=${padding.getPropertyValue('padding-left')}`;

            this.backdropElement.style.marginTop = `+=${border.getPropertyValue('border-top-width')}`;
            this.backdropElement.style.marginRight = `+=${border.getPropertyValue('border-right-width')}`;
            this.backdropElement.style.marginBottom = `+=${border.getPropertyValue('border-bottom-width')}`;
            this.backdropElement.style.marginLeft = `+=${border.getPropertyValue('border-left-width')}`;
        }
    };

    private fixIOS() {
        this.highlightsElement.style.paddingLeft = '3px';
        this.highlightsElement.style.paddingRight = '3px'
    }

    private handleInput() {
        let input = this?.targetElement?.value;
        let ranges = this.getRanges(input, this.highlightConfig);
        let unStaggeredRanges = this.removeStaggeredRanges(ranges);
        let boundaries: BoundaryType[] = this.getBoundaries(unStaggeredRanges);
        this.renderMarks(boundaries);
    };

    private getRanges(input: any, highlight: any): any[] {
        let type = this.getType(highlight);
        switch (type) {
            case 'array':
                return this.getArrayRanges(input, highlight);
            case 'function':
                return this.getFunctionRanges(input, highlight);
            case 'regexp':
                return this.getRegExpRanges(input, highlight);
            case 'string':
                return this.getStringRanges(input, highlight);
            case 'range':
                return this.getRangeRanges(input, highlight);
            case 'custom':
                return this.getCustomRanges(input, highlight);
            default:
                if (!highlight) {
                    return [];
                } else {
                    console.error('unrecognized highlight type');
                    throw new Error('unrecognized highlight type')
                }
        }
    };

    private getArrayRanges(input: string, arr: any): any[] {
        let ranges = arr.map((range: any) => this.getRanges(input, range));
        return [].concat(...ranges);
    };

    private getFunctionRanges(input: any, func: any): any[] {
        return this.getRanges(input, func(input));
    };

    private getRegExpRanges(input: any, regex: any): any[] {
        let ranges = [];
        let match;
        while ((match = regex.exec(input)) !== null) {
            ranges.push([match.index, match.index + match[0].length]);
            if (!regex.global) {
                break;
            }
        }
        return ranges;
    };

    private getStringRanges(input: any, str: any): any[] {
        let ranges = [];
        let inputLower = input.toLowerCase();
        let strLower = str.toLowerCase();
        let index = 0;
        while ((index = inputLower.indexOf(strLower, index)) !== -1) {
            ranges.push([index, index + strLower.length]);
            index += strLower.length;
        }
        return ranges;
    };

    private getRangeRanges(_: any, range: any) {
        return [range];
    };

    private getCustomRanges(input: any, custom: any) {
        let ranges = this.getRanges(input, custom.highlight);
        if (custom.className) {
            ranges.forEach(range => {
                if (range.className) {
                    range.className = `${custom.className} ${range.className}`;
                } else {
                    range.className = custom.className;
                }
            });
        }
        return ranges;
    };

    private removeStaggeredRanges(ranges: any): any {
        let unStaggeredRanges: any = [];
        ranges.forEach((range: any) => {
            let isStaggered = unStaggeredRanges.some((unStaggeredRange: any) => {
                let isStartInside = range[0] > unStaggeredRange[0] && range[0] < unStaggeredRange[1];
                let isStopInside = range[1] > unStaggeredRange[0] && range[1] < unStaggeredRange[1];
                return isStartInside !== isStopInside; // xor
            });
            if (!isStaggered) {
                unStaggeredRanges.push(range);
            }
        });
        return unStaggeredRanges;
    };

    private getBoundaries(ranges: any): BoundaryType[] {
        let boundaries: BoundaryType[] = [];
        ranges.forEach(function (range: any) {
            boundaries.push({
                type: 'start',
                index: range[0],
                className: range.className
            });
            boundaries.push({
                type: 'stop',
                index: range[1],
            });
        });

        this.sortBoundaries(boundaries);
        return boundaries;
    };

    private sortBoundaries(boundaries: BoundaryType[]) {
        boundaries.sort((a: BoundaryType, b: BoundaryType) => {
            if (a.index !== b.index) {
                return b.index - a.index;
            } else if (a.type === 'stop' && b.type === 'start') {
                return 1;
            } else if (a.type === 'start' && b.type === 'stop') {
                return -1;
            } else {
                return 0;
            }
        });
    };

    private renderMarks(boundaries: BoundaryType[]) {
        let input = this?.targetElement?.value;
        if (input) {
            boundaries.forEach((boundary: BoundaryType, index: number) => {
                let markup = (() => {
                    if (boundary.type === 'start') {
                        return '{{hwt-mark-start|' + index + '}}';
                    } else {
                        return '{{hwt-mark-stop}}';
                    }
                })()

                input = (input ?? '').slice(0, boundary.index) + markup + (input ?? '').slice(boundary.index);
            });

            input = input.replace(/\n({{hwt-mark-stop}})?$/, '\n\n$1');
            input = input.replace(/</g, '&lt;').replace(/>/g, '&gt;');

            if (this.userBrowser === 'ie') {
                input = input.replace(/ /g, ' <wbr>');
            }

            input = input.replace(/{{hwt-mark-start\|(\d+)}}/g, function (match, submatch) {
                const className = boundaries[+submatch].className;
                if (className) {
                    return '<mark class="' + className + '">';
                } else {
                    return '<mark>';
                }
            });

            input = input.replace(/{{hwt-mark-stop}}/g, '</mark>');

            this.highlightsElement.innerHTML = input;
        }
    };

    private handleScroll() {
        if (this.backdropElement && this.targetElement) {
            this.backdropElement.scrollTop = this.targetElement.scrollTop;

            let scrollLeft = this.targetElement.scrollLeft;
            this.backdropElement.style.transform = (scrollLeft > 0) ? `translateX(${-scrollLeft}px)` : '';
        }
    };

    private blockContainerScroll() {
        if (this.containerElement) {
            this.containerElement.scrollLeft = 0;
        }
    };

    destroy() {
        this.backdropElement?.remove();
        this?.targetElement?.parentElement?.removeChild(this?.targetElement); // targetElement을 부모에서 제거
    };
}
